// Quasar
// High-powed LED with an XBee
// http://lucaseckels.com/quasar
#include <WProgram.h>
#include <EEPROM.h>

#define MAX_LINE_LENGTH 64
#include "AtCommands.h"

#define RED_PIN 10
#define GREEN_PIN 9
#define BLUE_PIN 11

#define LIGHT_PIN 0
#define POWER_PIN 15
#define GND_PIN 16

#define STEPS 256

#define COLOR_TYPE 'C'
#define LOOP_TYPE 'L'
#define STOP_TYPE 'S'
#define MAX_STEPS 16
#define FIRST_PARAM() strtok(param, ",")
#define NEXT_PARAM() strtok(NULL, ",")

#define MAGIC "QS0"
#define MAGIC_LENGTH 3
#define ENABLE_DAMP_DEFAULT false

struct Settings {
  char *magic[MAGIC_LENGTH];
  bool enableDamp;
  int val0;
  byte damp0;
  int val1;
  byte damp1;
} settings;

struct Color {
  byte red;
  byte green;
  byte blue;
  int time_ms;
};

struct Loop {
  byte destination;
  byte count;
};

union Contents {
  Color color;
  Loop loop;
};

struct Step {
  byte type;
  Contents contents;
};

void readEeprom(void *buffer, size_t length) {
  byte *b = (byte*)buffer;
  for (size_t i = 0; i < length; ++i) {
    b[i] = EEPROM.read(i);
  }
}

void writeEeprom(void *buffer, size_t length) {
  byte *b = (byte*)buffer;
  for (size_t i = 0; i < length; ++i) {
    EEPROM.write(i, b[i]);
  }
}

void reset() {
  memcpy(settings.magic, MAGIC, MAGIC_LENGTH);
  settings.enableDamp = ENABLE_DAMP_DEFAULT;
  settings.val0 = 0;
  settings.damp0 = 0;
  settings.val1 = 0;
  settings.damp1 = 0;
}

Step steps[MAX_STEPS];
int current_step = 0;

bool running = false;
int sensor = 0;

void runSteps() {
  Serial.print("Starting steps\r\n");
  int red = 0;
  int green = 0;
  int blue = 0;
  running = true;
  sensor = analogRead(LIGHT_PIN);
  
  int i = 0;
  while (running) {
    switch (steps[i].type) {
      case COLOR_TYPE:
        doBlend(red, steps[i].contents.color.red,
                green, steps[i].contents.color.green,
                blue, steps[i].contents.color.blue,
                steps[i].contents.color.time_ms);
        red = steps[i].contents.color.red;
        green = steps[i].contents.color.green;
        blue = steps[i].contents.color.blue;
        ++i;
        break;
      case LOOP_TYPE:
        if (steps[i].contents.loop.count--) {
          i = steps[i].contents.loop.destination;
        } else {
          ++i;
        }
        break;
      case STOP_TYPE:
        running = false;
        break;
    }
  }
  analogWrite(RED_PIN, 0);
  analogWrite(GREEN_PIN, 0);
  analogWrite(BLUE_PIN, 0);
  Serial.print("Finished steps\r\n");
}
  

void setup() {
  Serial.begin(9600);
  Serial.print("Quasar starting\r\n");
  pinMode(RED_PIN, OUTPUT);
  pinMode(GREEN_PIN, OUTPUT);
  pinMode(BLUE_PIN, OUTPUT);
  
  pinMode(POWER_PIN, OUTPUT);
  pinMode(GND_PIN, OUTPUT);
  digitalWrite(POWER_PIN, HIGH);
  digitalWrite(GND_PIN, LOW);
  
  readEeprom(&settings, sizeof(Settings));
  if (memcmp(settings.magic, MAGIC, MAGIC_LENGTH)) {
    reset();
  }
}

static void atCommandHandlerQuasar();

void doBlend(int startRed, int stopRed,
             int startGreen, int stopGreen,
             int startBlue, int stopBlue,
             int time_ms) {
  if (time_ms == 0) {
    analogWrite(RED_PIN, stopRed);
    analogWrite(GREEN_PIN, stopGreen);
    analogWrite(BLUE_PIN, stopBlue);
    return;
  }
  
  int delay_us = 0;
  int delay_ms = time_ms / STEPS;
  if (delay_ms < 10) {
    delay_ms = 0;
    long time_us = time_ms * 1000L;
    delay_us = time_us / STEPS;
  }
  for (int i = 0; i < STEPS; ++i) {
    atCommandHandlerQuasar();
    int red = damp(interpolate(startRed, stopRed, i, STEPS));
    int green = damp(interpolate(startGreen, stopGreen, i, STEPS));
    int blue = damp(interpolate(startBlue, stopBlue, i, STEPS));
    analogWrite(RED_PIN, red);
    analogWrite(GREEN_PIN, green);
    analogWrite(BLUE_PIN, blue);
    if (delay_ms) {
      delay(delay_ms);
    } else {
      delayMicroseconds(delay_us);
    }
  }    
}

void SetColors(const char *command, char *param, byte length) {
  int red = atoi(strtok(param, ","));
  int green = atoi(strtok(NULL, ","));
  int blue = atoi(strtok(NULL, ","));
 
  analogWrite(RED_PIN, red);
  analogWrite(GREEN_PIN, green);
  analogWrite(BLUE_PIN, blue);
  delay(5000);

  atCommandOk();
}

void SayHi(const char *command, char *param, byte length) {
  Serial.print("Hi!\r\n");
  atCommandOk();
}

void SetStep(const char *command, char *param, byte length) {
  running = false;
  
  int step = atoi(FIRST_PARAM());
  steps[step].type = NEXT_PARAM()[0];
  switch (steps[step].type) {
    case COLOR_TYPE:
      steps[step].contents.color.red = atoi(NEXT_PARAM());
      steps[step].contents.color.green = atoi(NEXT_PARAM());
      steps[step].contents.color.blue = atoi(NEXT_PARAM());
      steps[step].contents.color.time_ms = atoi(NEXT_PARAM());
      break;
    case LOOP_TYPE:
      steps[step].contents.loop.destination = atoi(NEXT_PARAM());
      steps[step].contents.loop.count = atoi(NEXT_PARAM());
  }
  atCommandOk();
}

void StartRunning(const char *command, char *param, byte length) {
  running = true;
  atCommandOk();
}

void StopRunning(const char *command, char *param, byte length) {
  running = false;
  atCommandOk();
}

void DumpSteps(const char *command, char *param, byte length) {
  for (int i = 0; i < MAX_STEPS; ++i) {
    Serial.print("Step ");
    Serial.print(i);
    Serial.print("\r\n  Type: ");
    Serial.print(steps[i].type);
    switch (steps[i].type) {
      case COLOR_TYPE:
        Serial.print("\r\n  Red: ");
        Serial.print(steps[i].contents.color.red, DEC);
        Serial.print("\r\n  Green: ");
        Serial.print(steps[i].contents.color.green, DEC);
        Serial.print("\r\n  Blue: ");
        Serial.print(steps[i].contents.color.blue, DEC);
        Serial.print("\r\n  time_ms: ");
        Serial.print(steps[i].contents.color.time_ms, DEC);
        break;
      case LOOP_TYPE:
        Serial.print("\r\n  Dest: ");
        Serial.print(steps[i].contents.loop.destination, DEC);
        Serial.print("\r\n  Count: ");
        Serial.print(steps[i].contents.loop.count, DEC);
        break;
    }
    Serial.print("\r\n");
    delay(10);
  }  
  atCommandOk();
}

void LightRead(const char *command, char *param, int length) {
  int val = analogRead(LIGHT_PIN);
  Serial.print("Light value: ");
  Serial.print(val);
  Serial.print("\r\n");
}

void CommandReset(const char *command, char *buffer, byte length) {
  reset();
  atCommandOk();
}

void CommandWrite(const char *command, char *buffer, byte length) {
  writeEeprom(&settings, sizeof(Settings));
  atCommandOk();
}

void CommandPrintSettings(const char *command, char *buffer, byte length) {
  Serial.print("Damp enabled: ");
  Serial.print(settings.enableDamp);
  Serial.print("\r\nVal 0: ");
  Serial.print(settings.val0);
  Serial.print("\r\nDamp 0: ");
  Serial.print(settings.damp0, DEC);
  Serial.print("\r\nVal 1: ");
  Serial.print(settings.val1);
  Serial.print("\r\nDamp 1: ");
  Serial.print(settings.damp1, DEC);
  Serial.print("\r\n");
  atCommandOk();
}

void CommandSetSettings(const char *command, char *buffer, byte length) {
  int value = atoi(buffer);
  if (!strcmp(command, "V0")) {
    settings.val0 = value;
  } else if (!strcmp(command, "D0")) {
    settings.damp0 = value;
  } else if (!strcmp(command, "V1")) {
    settings.val1 = value;
  } else if (!strcmp(command, "D1")) {
    settings.damp1 = value;
  } else if (!strcmp(command, "ED")) {
    settings.enableDamp = value;
  }
  atCommandOk();
}

AT_COMMAND_MAP_BEGIN(Quasar)
AT_COMMAND_MAP_ENTRY("SE", SetColors)
AT_COMMAND_MAP_ENTRY("HI", SayHi)
AT_COMMAND_MAP_ENTRY("SS", SetStep)
AT_COMMAND_MAP_ENTRY("SR", StartRunning)
AT_COMMAND_MAP_ENTRY("ST", StopRunning)
AT_COMMAND_MAP_ENTRY("DS", DumpSteps)
AT_COMMAND_MAP_ENTRY("LR", LightRead)
AT_COMMAND_MAP_ENTRY("RE", CommandReset)
AT_COMMAND_MAP_ENTRY("WR", CommandWrite)
AT_COMMAND_MAP_ENTRY("PR", CommandPrintSettings)
AT_COMMAND_MAP_ENTRY("ED", CommandSetSettings)
AT_COMMAND_MAP_ENTRY("V0", CommandSetSettings)
AT_COMMAND_MAP_ENTRY("D0", CommandSetSettings)
AT_COMMAND_MAP_ENTRY("V1", CommandSetSettings)
AT_COMMAND_MAP_ENTRY("D1", CommandSetSettings)
AT_COMMAND_MAP_END(Quasar)

void loop() {
  while (!running) {
    atCommandHandlerQuasar();
  }
  runSteps();
}

int interpolate(int start, int target, int position, int positionSteps) {
  float relativePos = (position * 1.0) / positionSteps;
  int diff = target - start;
  int val = start + diff * relativePos;
  return val;
}

int damp(int value) {
  if (!settings.enableDamp) {
    return value;
  }
  byte dampAmount = 256;
  if (sensor >= settings.val1) {
    dampAmount = settings.damp1;
  } else if (sensor <= settings.val0) {
    dampAmount = settings.damp0;
  } else {
    int span = settings.val1 - settings.val0;
    dampAmount = interpolate(settings.damp1, settings.damp0, settings.val1 - sensor, span);
  }
  float dampFloat = dampAmount / 256.0;
  return value * dampFloat;
}
    
  


